@@ -1,22 +1,22 @@
 /* Miscellaneous support for test programs.
 
-Copyright 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008 Free Software Foundation, Inc.
+Copyright 2001, 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009 Free Software Foundation, Inc.
 Contributed by the Arenaire and Cacao projects, INRIA.
 
-This file is part of the MPFR Library.
+This file is part of the GNU MPFR Library.
 
-The MPFR Library is free software; you can redistribute it and/or modify
+The GNU MPFR Library is free software; you can redistribute it and/or modify
 it under the terms of the GNU Lesser General Public License as published by
 the Free Software Foundation; either version 2.1 of the License, or (at your
 option) any later version.
 
-The MPFR Library is distributed in the hope that it will be useful, but
+The GNU MPFR Library is distributed in the hope that it will be useful, but
 WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
 or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU Lesser General Public
 License for more details.
 
 You should have received a copy of the GNU Lesser General Public License
-along with the MPFR Library; see the file COPYING.LIB.  If not, write to
+along with the GNU MPFR Library; see the file COPYING.LIB.  If not, write to
 the Free Software Foundation, Inc., 51 Franklin St, Fifth Floor, Boston,
 MA 02110-1301, USA. */
 
@@ -30,21 +30,24 @@ MA 02110-1301, USA. */
 #include <stdlib.h>
 #include <string.h>
 #include <float.h>
+#include <errno.h>
 
-#if HAVE_SETLOCALE
+#ifdef HAVE_LOCALE_H
 #include <locale.h>
 #endif
 
-#if TIME_WITH_SYS_TIME
+#ifdef TIME_WITH_SYS_TIME
 # include <sys/time.h>  /* for struct timeval */
 # include <time.h>
-#elif HAVE_SYS_TIME_H
+#elif defined HAVE_SYS_TIME_H
 #  include <sys/time.h>
 #else
 #  include <time.h>
 #endif
 
-#if HAVE_SYS_FPU_H
+/* <sys/fpu.h> is needed to have union fpc_csr defined under IRIX64
+   (see below). Let's include it only if need be. */
+#if defined HAVE_SYS_FPU_H && defined HAVE_FPC_CSR
 # include <sys/fpu.h>
 #endif
 
@@ -68,6 +71,14 @@ MA 02110-1301, USA. */
  *   make check CFLAGS="-g -O2 -ffloat-store -DMPFR_FPU_PREC=_FPU_SINGLE"
  *
  * i.e. just add -DMPFR_FPU_PREC=... to the CFLAGS found in Makefile.
+ *
+ * Notes:
+ *   + SSE2 (used to implement double's on x86_64, and possibly on x86
+ *     too, depending on the compiler configuration and flags) is not
+ *     affected by the dynamic precision.
+ *   + When the FPU is set to single precision, the behavior of MPFR
+ *     functions that have a native floating-point type (float, double,
+ *     long double) as argument or return value is not guaranteed.
  */
 
 #include <fpu_control.h>
@@ -133,7 +144,7 @@ tests_start_mpfr (void)
   /* don't buffer, so output is not lost if a test causes a segv etc */
   setbuf (stdout, NULL);
 
-#if HAVE_SETLOCALE
+#if defined HAVE_LOCALE_H && defined HAVE_SETLOCALE
   /* Added on 2005-07-09. This allows to test MPFR under various
      locales. New bugs will probably be found, in particular with
      LC_ALL="tr_TR.ISO8859-9" because of the i/I character... */
@@ -239,7 +250,7 @@ tests_rand_start (void)
         }
       else
         {
-#if HAVE_GETTIMEOFDAY
+#ifdef HAVE_GETTIMEOFDAY
           struct timeval  tv;
           gettimeofday (&tv, NULL);
           seed = tv.tv_sec + tv.tv_usec;
@@ -266,10 +277,10 @@ tests_rand_end (void)
 /* initialization function for tests using the hardware floats
    Not very useful now. */
 void
-mpfr_test_init ()
+mpfr_test_init (void)
 {
   double d;
-#if HAVE_FPC_CSR
+#ifdef HAVE_FPC_CSR
   /* to get denormalized numbers on IRIX64 */
   union fpc_csr exp;
 
@@ -364,6 +375,7 @@ dbl (double m, int e)
   return m;
 }
 
+/* Warning: NaN values cannot be distinguished if MPFR_NANISNAN is defined. */
 int
 Isnan (double d)
 {
@@ -426,7 +438,8 @@ src_fopen (const char *filename, const char *mode)
 
   if (srcdir == NULL)
     return fopen (filename, mode);
-  buffer = (char*) malloc (strlen (filename) + strlen (srcdir) + 2);
+  buffer =
+    (char*) (*__gmp_allocate_func) (strlen (filename) + strlen (srcdir) + 2);
   if (buffer == NULL)
     {
       printf ("src_fopen: failed to alloc memory)\n");
@@ -434,7 +447,7 @@ src_fopen (const char *filename, const char *mode)
     }
   sprintf (buffer, "%s/%s", srcdir, filename);
   f = fopen (buffer, mode);
-  free (buffer);
+  (*__gmp_free_func) (buffer, strlen (filename) + strlen (srcdir) + 2);
   return f;
 }
 
@@ -458,16 +471,112 @@ set_emax (mp_exp_t exponent)
     }
 }
 
+/* pos is 512 times the proportion of negative numbers.
+   If pos=256, half of the numbers are negative.
+   If pos=0, all generated numbers are positive.
+*/
 void
-tests_default_random (mpfr_ptr x)
+tests_default_random (mpfr_ptr x, int pos, mp_exp_t emin, mp_exp_t emax)
 {
-  mpfr_random (x);
-  if (randlimb () & 1)
-    mpfr_mul_2si (x, x, (int) (randlimb () % 512) - 256, GMP_RNDN);
-  if (randlimb () & 1)
+  MPFR_ASSERTN (emin <= emax);
+  MPFR_ASSERTN (emin >= MPFR_EMIN_MIN);
+  MPFR_ASSERTN (emax <= MPFR_EMAX_MAX);
+  /* but it isn't required that emin and emax are in the current
+     exponent range (see below), so that underflow/overflow checks
+     can be done on 64-bit machines. */
+
+  mpfr_urandomb (x, RANDS);
+  if (MPFR_IS_PURE_FP (x) && (emin >= 1 || (randlimb () & 1)))
+    {
+      mp_exp_t e;
+      e = MPFR_GET_EXP (x) +
+        (emin + (long) (randlimb () % (emax - emin + 1)));
+      /* Note: There should be no overflow here because both terms are
+         between MPFR_EMIN_MIN and MPFR_EMAX_MAX, but the sum e isn't
+         necessarily between MPFR_EMIN_MIN and MPFR_EMAX_MAX. */
+      if (mpfr_set_exp (x, e))
+        {
+          /* The random number doesn't fit in the current exponent range.
+             In this case, test the function in the extended exponent range,
+             which should be restored by the caller. */
+          mpfr_set_emin (MPFR_EMIN_MIN);
+          mpfr_set_emax (MPFR_EMAX_MAX);
+          mpfr_set_exp (x, e);
+        }
+    }
+  if (randlimb () % 512 < pos)
     mpfr_neg (x, x, GMP_RNDN);
 }
 
+/* The test_one argument is a boolean. If it is true, then the function
+   is tested in only one rounding mode (the one provided in rnd) and the
+   variable rndnext is not used (due to the break). If it is false, then
+   the function is tested in the 4 rounding modes, and rnd must initially
+   be GMP_RNDZ; thus rndnext will be initialized in the first iteration.
+   gcc may give a warning about rndnext, but this is an easy and correct
+   way to implement a simple queue for the rounding modes.
+   As examples of use, see the calls to test4rm from the data_check and
+   bad_cases functions. */
+static void
+test4rm (int (*fct) (FLIST), mpfr_srcptr x, mpfr_ptr y, mpfr_ptr z,
+         mp_rnd_t rnd, int test_one, char *name)
+{
+  mp_prec_t yprec = MPFR_PREC (y);
+  mp_rnd_t rndnext = GMP_RND_MAX;  /* means uninitialized */
+
+  MPFR_ASSERTN (test_one || rnd == GMP_RNDZ);
+  mpfr_set_prec (z, yprec);
+  while (1)
+    {
+      MPFR_ASSERTN (rnd != GMP_RND_MAX);
+      fct (z, x, rnd);
+      if (! mpfr_equal_p (y, z))
+        {
+          printf ("Error for %s with xprec=%lu, yprec=%lu, rnd=%s\nx = ",
+                  name, (unsigned long) MPFR_PREC (x), (unsigned long) yprec,
+                  mpfr_print_rnd_mode (rnd));
+          mpfr_out_str (stdout, 16, 0, x, GMP_RNDN);
+          printf ("\nexpected ");
+          mpfr_out_str (stdout, 16, 0, y, GMP_RNDN);
+          printf ("\ngot      ");
+          mpfr_out_str (stdout, 16, 0, z, GMP_RNDN);
+          printf ("\n");
+          exit (1);
+        }
+      if (test_one || rnd == GMP_RNDN)
+        break;
+      if (rnd == GMP_RNDZ)
+        {
+          if (MPFR_IS_NEG (y))
+            {
+              rnd = GMP_RNDU;
+              rndnext = GMP_RNDD;
+            }
+          else
+            {
+              rnd = GMP_RNDD;
+              rndnext = GMP_RNDU;
+            }
+        }
+      else
+        {
+          rnd = rndnext;
+          if (rndnext != GMP_RNDN)
+            {
+              rndnext = GMP_RNDN;
+              mpfr_nexttoinf (y);
+            }
+          else
+            {
+              if (yprec == MPFR_PREC_MIN)
+                break;
+              mpfr_prec_round (y, --yprec, GMP_RNDZ);
+              mpfr_set_prec (z, yprec);
+            }
+        }
+    }
+}
+
 /* Check data in file f for function foo, with name 'name'.
    Each line consists of the file f one:
 
@@ -488,19 +597,19 @@ tests_default_random (mpfr_ptr x)
    directed rounding mode.
  */
 void
-data_check (char *f, int (*foo) (), char *name)
+data_check (char *f, int (*foo) (FLIST), char *name)
 {
   FILE *fp;
   mp_prec_t xprec, yprec;
   mpfr_t x, y, z;
-  mp_rnd_t rnd, rndnext;
+  mp_rnd_t rnd;
   char r;
   int c;
 
   fp = fopen (f, "r");
   if (fp == NULL)
     {
-      char *v = MPFR_VERSION_STRING;
+      char *v = (char *) MPFR_VERSION_STRING;
 
       /* In the '-dev' versions, assume that the data file exists and
          return an error if the file cannot be opened to make sure
@@ -522,10 +631,29 @@ data_check (char *f, int (*foo) (), char *name)
 
   while (!feof (fp))
     {
-      fscanf (fp, " ");  /* skip whitespace, for consistency */
-      c = getc (fp);
-      if (c == EOF)
-        break;
+      /* skip whitespace, for consistency */
+      if (fscanf (fp, " ") == EOF)
+        {
+          if (ferror (fp))
+            {
+              perror ("data_check");
+              exit (1);
+            }
+          else
+            break;  /* end of file */
+        }
+
+      if ((c = getc (fp)) == EOF)
+        {
+          if (ferror (fp))
+            {
+              perror ("data_check");
+              exit (1);
+            }
+          else
+            break;  /* end of file */
+        }
+
       if (c == '#') /* comment: read entire line */
         {
           do
@@ -537,11 +665,19 @@ data_check (char *f, int (*foo) (), char *name)
       else
         {
           ungetc (c, fp);
-          if (fscanf (fp, "%lu %lu %c", &xprec, &yprec, &r) != 3)
+
+          c = fscanf (fp, "%lu %lu %c", &xprec, &yprec, &r);
+          if (c == EOF)
+            {
+              perror ("data_check");
+              exit (1);
+            }
+          else if (c != 3)
             {
               printf ("Error: corrupted line in file '%s'\n", f);
               exit (1);
             }
+
           switch (r)
             {
             case 'n':
@@ -563,7 +699,6 @@ data_check (char *f, int (*foo) (), char *name)
             }
           mpfr_set_prec (x, xprec);
           mpfr_set_prec (y, yprec);
-          mpfr_set_prec (z, yprec);
           if (mpfr_inp_str (x, fp, 0, GMP_RNDN) == 0)
             {
               printf ("Error: corrupted argument in file '%s'\n", f);
@@ -580,55 +715,12 @@ data_check (char *f, int (*foo) (), char *name)
               exit (1);
             }
           /* Skip whitespace, in particular at the end of the file. */
-          fscanf (fp, " ");
-
-          while (1)
+          if (fscanf (fp, " ") == EOF && ferror (fp))
             {
-              foo (z, x, rnd);
-              if (! mpfr_equal_p (y, z))
-                {
-                  printf ("Error for %s with xprec=%ld, yprec=%ld, rnd=%s\nx=",
-                          name, xprec, yprec, mpfr_print_rnd_mode (rnd));
-                  mpfr_out_str (stdout, 16, 0, x, GMP_RNDN);
-                  printf ("\nexpected ");
-                  mpfr_out_str (stdout, 16, 0, y, GMP_RNDN);
-                  printf ("\ngot      ");
-                  mpfr_out_str (stdout, 16, 0, z, GMP_RNDN);
-                  printf ("\n");
-                  exit (1);
-                }
-              if (r != 'Z' || rnd == GMP_RNDN)
-                break;
-              if (rnd == GMP_RNDZ)
-                {
-                  if (MPFR_IS_NEG (y))
-                    {
-                      rnd = GMP_RNDU;
-                      rndnext = GMP_RNDD;
-                    }
-                  else
-                    {
-                      rnd = GMP_RNDD;
-                      rndnext = GMP_RNDU;
-                    }
-                }
-              else
-                {
-                  rnd = rndnext;
-                  if (rndnext != GMP_RNDN)
-                    {
-                      rndnext = GMP_RNDN;
-                      mpfr_nexttoinf (y);
-                    }
-                  else
-                    {
-                      if (yprec == MPFR_PREC_MIN)
-                        break;
-                      mpfr_prec_round (y, --yprec, GMP_RNDZ);
-                      mpfr_set_prec (z, yprec);
-                    }
-                }
+              perror ("data_check");
+              exit (1);
             }
+          test4rm (foo, x, y, z, rnd, r != 'Z', name);
         }
     }
 
@@ -638,3 +730,132 @@ data_check (char *f, int (*foo) (), char *name)
 
   fclose (fp);
 }
+
+/* Test n random bad cases. A precision py in [pymin,pymax] and
+ * a number y of precision py are chosen randomly. One computes
+ * x = inv(y) in precision px = py + psup (rounded to nearest).
+ * Then (in general), y is a bad case for fct in precision py (in
+ * the directed rounding modes, but also in the rounding-to-nearest
+ * mode for some lower precision: see data_check).
+ * fct, inv, name: data related to the function.
+ * pos, emin, emax: arguments for tests_default_random.
+ */
+void
+bad_cases (int (*fct)(FLIST), int (*inv)(FLIST), char *name,
+           int pos, mp_exp_t emin, mp_exp_t emax,
+           mp_prec_t pymin, mp_prec_t pymax, mp_prec_t psup,
+           int n)
+{
+  mpfr_t x, y, z;
+  char *dbgenv;
+  int i, dbg;
+  mp_exp_t old_emin, old_emax;
+
+  old_emin = mpfr_get_emin ();
+  old_emax = mpfr_get_emax ();
+
+  dbgenv = getenv ("MPFR_DEBUG_BADCASES");
+  dbg = dbgenv != 0 ? atoi (dbgenv) : 0;  /* debug level */
+  mpfr_inits (x, y, z, (mpfr_ptr) 0);
+  for (i = 0; i < n; i++)
+    {
+      mp_prec_t px, py, pz;
+      int inex;
+
+      if (dbg)
+        printf ("bad_cases: i = %d\n", i);
+      py = pymin + (randlimb () % (pymax - pymin + 1));
+      mpfr_set_prec (y, py);
+      tests_default_random (y, pos, emin, emax);
+      if (dbg)
+        {
+          printf ("bad_cases: yprec =%4ld, y = ", (long) py);
+          mpfr_out_str (stdout, 16, 0, y, GMP_RNDN);
+          printf ("\n");
+        }
+      px = py + psup;
+      mpfr_set_prec (x, px);
+      mpfr_clear_flags ();
+      inv (x, y, GMP_RNDN);
+      if (mpfr_nanflag_p () || mpfr_overflow_p () || mpfr_underflow_p ())
+        {
+          if (dbg)
+            printf ("bad_cases: no normal inverse\n");
+          goto next_i;
+        }
+      if (dbg > 1)
+        {
+          printf ("bad_cases: x = ");
+          mpfr_out_str (stdout, 16, 0, x, GMP_RNDN);
+          printf ("\n");
+        }
+      pz = px;
+      do
+        {
+          pz += 32;
+          mpfr_set_prec (z, pz);
+          if (fct (z, x, GMP_RNDN) == 0)
+            {
+              if (dbg)
+                printf ("bad_cases: exact case\n");
+              goto next_i;
+            }
+          if (dbg)
+            {
+              if (dbg > 1)
+                {
+                  printf ("bad_cases: %s(x) ~= ", name);
+                  mpfr_out_str (stdout, 16, 0, z, GMP_RNDN);
+                }
+              else
+                {
+                  printf ("bad_cases:   [GMP_RNDZ]  ~= ");
+                  mpfr_out_str (stdout, 16, 40, z, GMP_RNDZ);
+                }
+              printf ("\n");
+            }
+          inex = mpfr_prec_round (z, py, GMP_RNDN);
+          if (mpfr_nanflag_p () || mpfr_overflow_p () || mpfr_underflow_p ()
+              || ! mpfr_equal_p (z, y))
+            {
+              if (dbg)
+                printf ("bad_cases: inverse doesn't match\n");
+              goto next_i;
+            }
+        }
+      while (inex == 0);
+      /* We really have a bad case. */
+      do
+        py--;
+      while (py >= MPFR_PREC_MIN && mpfr_prec_round (z, py, GMP_RNDZ) == 0);
+      py++;
+      /* py is now the smallest output precision such that we have
+         a bad case in the directed rounding modes. */
+      if (mpfr_prec_round (y, py, GMP_RNDZ) != 0)
+        {
+          printf ("Internal error for i = %d\n", i);
+          exit (1);
+        }
+      if ((inex > 0 && MPFR_IS_POS (z)) ||
+          (inex < 0 && MPFR_IS_NEG (z)))
+        {
+          mpfr_nexttozero (y);
+          if (mpfr_zero_p (y))
+            goto next_i;
+        }
+      if (dbg)
+        {
+          printf ("bad_cases: yprec =%4ld, y = ", (long) py);
+          mpfr_out_str (stdout, 16, 0, y, GMP_RNDN);
+          printf ("\n");
+        }
+      /* Note: y is now the expected result rounded towards zero. */
+      test4rm (fct, x, y, z, GMP_RNDZ, 0, name);
+    next_i:
+      /* In case the exponent range has been changed by
+         tests_default_random()... */
+      mpfr_set_emin (old_emin);
+      mpfr_set_emax (old_emax);
+    }
+  mpfr_clears (x, y, z, (mpfr_ptr) 0);
+}
